/*
 * rshaper.c
 *
 * Copyright (c) 1998-2001  Alessandro Rubini <rubini@linux.it>
 * Copyright (c) 2001       Rodolfo Giometti <giometti@linux.it>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#undef PDEBUG             /* undef it, just in case */
#ifdef RSHAPER_DEBUG
#  ifdef __KERNEL__
     /* This one if debugging is on, and kernel space */
#    define PDEBUG(fmt, args...) printk( KERN_DEBUG "rshaper: " fmt, ## args)
#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#  define PDEBUG(fmt, args...) /* not debugging: nothing */
#endif

#undef PDEBUGG
#define PDEBUGG(fmt, args...) /* nothing: it's a placeholder */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/interrupt.h> /* mark_bh */

#include <linux/netdevice.h>   /* struct device, and other headers */
#include <linux/etherdevice.h> /* eth_type_trans */
#include <linux/ip.h>          /* struct iphdr */
#include <linux/tcp.h>         /* struct tcphdr */
#include <linux/skbuff.h>

#include "rshaper.h" 

#include "sysdep.h"

#ifdef LINUX_24
  /* Load time parameter */
  static int mode = 1; /* Incoming */
  MODULE_PARM(mode, "i");
#else
  /* external hook, local implementation */
  extern int (*net_shaper_rx_hook)(struct sk_buff *skb); /* in the eth driver */
  int rshaper_external_hook(struct sk_buff *skb);
#endif

/* ------------------------------------------------ Data Structures */

/* Each queued packet */
struct rshaper_packet {
    struct sk_buff *skb;
    int (*okfn)(struct sk_buff *);              /* reinject function */
    unsigned long timestamp;
    struct rshaper_packet *next;
};

/* Each queue */
struct rshaper_queue {
    struct rshaper_packet *head; 	/* first packet, to dequeue */
    struct rshaper_packet *tail; 	/* last packet, to enqueue */
    struct timer_list timer;   		/* use this to dequeue */
    unsigned long nexttick;     	/* foreseen next tick */
    int approx;                  	/* leftover bytes */
    spinlock_t lock;			/* mutex lock */
};

/* Each host or host group */
struct rshaper_host {
    u32 ip;
    u32 mask;
    struct rshaper_queue *tx_queue;
    struct rshaper_queue *rx_queue;
    struct rshaper_host *next;
    int bytes_per_tick;				/* the bandwidth in B/tick */
    int timeout; /* measured in jiffies */
};    

struct rshaper_host *first_host;
spinlock_t rshaper_host_lock;


/* ------------------------------------------------ Queue functions */

/* this is entered with the queue lock held */
int enqueue_packet(struct sk_buff *skb,
                              int (*okfn)(struct sk_buff *),
                              struct rshaper_queue *q,
                              int bytes_per_tick)
{
    struct rshaper_packet *pkt;
    int size = skb->tail - skb->data;
    int timerflag=0;

    /* if the `next tick time' is expired, transmit it now */
    if (q->nexttick <= jiffies) {
        PDEBUG("enqueue: now\n");
        if (q->nexttick < jiffies)
            q->approx = 0;
        size += q->approx;
        q->nexttick = jiffies + size / bytes_per_tick;
        q->approx = size % bytes_per_tick;
	rshaper_accept_fun(skb); /* for 2.0 and 2.2 */
        return NF_ACCEPT; /* for 2.4 */
    }

    /* allocate a structure */
    pkt = kmalloc(sizeof(struct rshaper_packet), GFP_ATOMIC);
    if (!pkt) {
	printk(KERN_INFO "rshaper: allocation failed\n");
	kfree_skb(skb);
        return NF_STOLEN; /* ENOMEM */
    }

    /* fill it */
    pkt->skb = skb;
    pkt->okfn = okfn;
    pkt->next = NULL;
    pkt->timestamp = q->nexttick;

    PDEBUG("enqueue: %04i+%03i bytes, %3i tick, %3i left (delay %li j)\n",
           size, q->approx, (size+q->approx)/bytes_per_tick,
           (size+q->approx)%bytes_per_tick,
           q->nexttick - jiffies);
    size += q->approx;
    q->nexttick += size / bytes_per_tick;
    q->approx = size % bytes_per_tick;

    if (q->head) {
        PDEBUGG("updating tail (h %p, t %p)\n",q->head,q->tail);
        q->tail->next = pkt;
        q->tail = pkt;
    } else {
        PDEBUGG("registering the timer to run at %li\n",pkt->timestamp);
        q->head = q->tail = pkt;
        q->timer.expires = pkt->timestamp;
        timerflag++;
    }

    if (timerflag)
        add_timer(&q->timer);
    return NF_STOLEN;
}

/* This is called by the timer, so we must get the lock ourselves */
void dequeue_packet(unsigned long data)
{
    struct rshaper_queue *q = (struct rshaper_queue *)data;
    struct rshaper_packet *p;
    unsigned long flags;

    PDEBUGG("timer running: now %li\n",jiffies);

    spin_lock_irqsave(&q->lock, flags);

    /* consume expired packets */
    while ((p = q->head) && (p->timestamp <= jiffies)) {
        q->head = p->next;
        if (q->tail == p)
            q->tail = NULL;
	/* release the lock during reinjection, then reacquire it */
	spin_unlock_irqrestore(&q->lock, flags);
        p->okfn(p->skb);
        kfree(p);
	spin_lock_irqsave(&q->lock, flags);
    }

    /* if some more is there, schedule it */
    if (p) {
        q->timer.expires = p->timestamp;
        add_timer(&q->timer); /* won't be run before we return */
    }
    spin_unlock_irqrestore(&q->lock, flags);
}


/* ------------------------------------------------ Limit the queue */

#ifndef RSHAPER_LIMIT_KB
#  define RSHAPER_LIMIT_KB 512 /* half a meg per queue */
#endif

/* return a new timeout (a jiffy value) according to the max queue size */
int rshaper_retimeout(int timeout, int bytes_per_tick)
{
    int maxticks = RSHAPER_LIMIT_KB * 1024 / bytes_per_tick;
    if (maxticks < timeout) timeout = maxticks;
    return timeout;
}

/* ------------------------------------------------ Host Management */

/* this is called with the host lock held */
struct rshaper_host *rshaper_find_host (u32 ip, struct rshaper_host ***prevp)
{
    struct rshaper_host *host, **prev = &first_host;

    PDEBUGG("Finding host %d.%d.%d.%d\n", NIPQUAD(ip));
    while (*prev && ((*prev)->ip != ip))
        prev = &((*prev)->next);

    if (!*prev) {
        *prevp = NULL;
        return NULL;
    }
    host = *prev;
    *prevp = prev; 
    return host;
}

/* this is called with the host lock held */
int rshaper_add_host(u32 ip, u32 mask, int bytes_per_tick, int timeout)
{
    struct rshaper_host *host;
    
    host = kmalloc(sizeof(struct rshaper_host), GFP_KERNEL);
    if (!host) return -ENOMEM;

    PDEBUG("Adding host %d.%d.%d.%d, %i bpt, %i jiffies\n",
           NIPQUAD(ip), bytes_per_tick, timeout);

    host->ip = ip;
    host->mask = mask;
    host->bytes_per_tick = bytes_per_tick;
    host->timeout = timeout;
    host->tx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_KERNEL);
    host->rx_queue = kmalloc(sizeof(struct rshaper_queue), GFP_KERNEL);

    if (!host->tx_queue || !host->rx_queue) {
        if (host->tx_queue) kfree(host->tx_queue);
        if (host->rx_queue) kfree(host->rx_queue);
        kfree(host);
        return -ENOMEM;
    }
    memset(host->rx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->rx_queue->timer);
    host->rx_queue->timer.data = (unsigned long)host->rx_queue;
    host->rx_queue->timer.function = dequeue_packet;
    spin_lock_init(&host->rx_queue->lock);

    memset(host->tx_queue, 0, sizeof(struct rshaper_queue));
    init_timer(&host->tx_queue->timer);
    host->tx_queue->timer.data = (unsigned long)host->tx_queue;
    host->tx_queue->timer.function = dequeue_packet;
    spin_lock_init(&host->tx_queue->lock);

    host->next = first_host;
    first_host = host;
    PDEBUGG("Registered with tx %p, rx %p\n", host->tx_queue, host->rx_queue);
    return 0;
    
}

/* this is called after removing the host from the list (host lock held) */
void rshaper_kfree_host(struct rshaper_host *host)
{
    struct rshaper_packet *packet, *nextp;

    del_timer_sync(&host->tx_queue->timer);
    del_timer_sync(&host->rx_queue->timer);

    /* no need to held the queue lock, as nobody will add/remove packets */
    for (packet = host->tx_queue->head; packet; ) {
        nextp = packet->next;
        PDEBUGG("free packet %p (next %p)\n", packet, nextp);
        packet->okfn(packet->skb);
        kfree(packet);
        packet = nextp;
    }
    kfree(host->tx_queue);
    
    for (packet = host->rx_queue->head; packet; ) {
        nextp = packet->next;
        PDEBUGG("free packet %p (next %p)\n", packet, nextp);
        packet->okfn(packet->skb);
        kfree(packet);
        packet = nextp;
    }
    kfree(host->rx_queue);
    PDEBUG("free host %p\n", host);
    kfree(host);
}


/* ------------------------------------------------ Main Hook */

unsigned int rshaper_rx_hook(unsigned int hooknum,
                             struct sk_buff **skb_p,
                             const struct net_device *in,
                             const struct net_device *out,
                             int (*okfn)(struct sk_buff *))
{
    struct sk_buff *skb = (*skb_p);

    /* no need to skip eth header, as this is after eth_type_trans() */
    struct iphdr *ih = (struct iphdr *)(skb->data);
    struct rshaper_host *ptr;
    u32 saddr = ih->saddr;
    u32 daddr = ih->daddr;
    unsigned long flags;
    int retval = NF_ACCEPT;

    PDEBUGG("packet: from %d.%d.%d.%d to %d.%d.%d.%d\n",
            NIPQUAD(ih->saddr), NIPQUAD(ih->daddr));

    /*
     * Ok, now we must check out our list of managed hosts, and delay
     * reception if found
     */
    for (ptr = first_host; ptr; ptr = ptr->next) {
        if (ptr->ip == (saddr & ptr->mask)) {
            PDEBUGG("Enqueue tx: %d.%d.%d.%d\n", NIPQUAD(saddr));
            if (ptr->tx_queue->nexttick <= jiffies + ptr->timeout) {
		spin_lock_irqsave(&ptr->tx_queue->lock, flags);
                retval =
		  enqueue_packet(skb, okfn, ptr->tx_queue, ptr->bytes_per_tick);
		spin_unlock_irqrestore(&ptr->tx_queue->lock, flags);
            } else {
                PDEBUG("Tx queue too long (%li+%i < %li, dropped\n",
		       jiffies, ptr->timeout, ptr->tx_queue->nexttick);
		rshaper_drop_fun(skb);
                retval = NF_DROP;
            }
            return retval;
        }
        if (ptr->ip == (daddr & ptr->mask))  {
            PDEBUGG("Enqueue rx: %d.%d.%d.%d\n", NIPQUAD(daddr));
            if (ptr->rx_queue->nexttick <= jiffies + ptr->timeout) {
		spin_lock_irqsave(&ptr->rx_queue->lock, flags);
                retval =
		  enqueue_packet(skb, okfn, ptr->rx_queue, ptr->bytes_per_tick);
		spin_unlock_irqrestore(&ptr->rx_queue->lock, flags);
            } else {
                PDEBUG("Rx queue too long (%li+%i < %li, dropped\n",
		       jiffies, ptr->timeout, ptr->rx_queue->nexttick);
		rshaper_drop_fun(skb);
                retval = NF_DROP;
            }
            return retval;
        }
    }

    /* no delay, receive it right now */
    rshaper_accept_fun(skb); /* for 2.0 and 2.2 */
    return retval; /* for 2.4 */
}

#ifdef LINUX_24
struct nf_hook_ops rshaper_rx_hook_ops = {   /* The pre-routing hook */
   hook :       rshaper_rx_hook,
   pf :         PF_INET,
   hooknum :    NF_IP_PRE_ROUTING,
};

struct nf_hook_ops rshaper_tx_hook_ops = {   /* The post-routing hook */
   hook :       rshaper_rx_hook,   /* Yes! It's the same as above! ;-p */
   pf :         PF_INET,
   hooknum :    NF_IP_POST_ROUTING,
};
#endif

/* ------------------------------------------------ User Cfg */

/*
 * There are several ways to talk to user space, devices, network interfaces,
 * /proc/entry points... Here I chose to use ioctl() on a network interface.
 *
 * SIOCDEVPRIVATE on rshaper is used to set host information,
 * SIOCDEVPRIVATE+1 is used to retrieve information.
 */
int rshaper_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct rshaper_userinfo info,
         *uptr = (struct rshaper_userinfo *)ifr->ifr_data;
    struct rshaper_host *host, **prev;
    unsigned long flags;
    int i, bw;

    if (!suser()) return -EPERM;

    /* use the verify_area() call if we are running 2.0 */
    if (!access_ok(VERIFY_READ, uptr, sizeof(*uptr)))
        return -EFAULT;
    copy_from_user(&info, uptr, sizeof(info));

    switch(cmd) {
      case SIOCRSHAPERSETINFO: /* set or clear one host */
	PDEBUG("ioctl: host %d.%d.%d.%d, count %i, to %i ms\n",
	       NIPQUAD(info.ip), (int)info.bytes_per_second, info.timeout);

	/* Some sanity checks */
	if (info.bytes_per_second && info.bytes_per_second<HZ) return -EINVAL;
	bw = (info.bytes_per_second+HZ/2)/HZ;
	if (info.timeout < 0 || info.timeout > 30000) return -EINVAL;
	info.timeout = info.timeout * HZ / 1000; /* ms to jiffies */

	spin_lock_irqsave(&rshaper_host_lock, flags);
        host = rshaper_find_host((u32)(info.ip), &prev);
        if (host) {
            if (info.bytes_per_second)  {
                /* turn byterate into bytes/tick, use nearest value */
                host->bytes_per_tick = bw;
                host->timeout = /* have a limit on the queue length */
		    rshaper_retimeout(info.timeout, bw);
            } else { /* remove host */
                *prev = host->next;
                rshaper_kfree_host(host);
            }
        }
        else if (info.bytes_per_second) {
            rshaper_add_host(info.ip, info.mask,
			     bw, rshaper_retimeout(info.timeout, bw));
        }
	spin_unlock_irqrestore(&rshaper_host_lock, flags);
        return 0;

      case SIOCRSHAPERGETINFO: /* return information about the host */
        if (!access_ok(VERIFY_WRITE, uptr, sizeof(*uptr)))
            return -EFAULT;

	spin_lock_irqsave(&rshaper_host_lock, flags);
        for (host = first_host, i=0; host && i < info.ip;
             host = host->next, i++)
            /* nothing */;
        if (host) {
	    info.ip = (__u32)(host->ip);
	    info.mask = (__u32)(host->mask);
	    info.bytes_per_second = HZ * host->bytes_per_tick;
	    PDEBUG("to %i\n", host->timeout);
	    info.timeout = host->timeout * 1000/HZ; /* milliseconds */
	}
	spin_unlock_irqrestore(&rshaper_host_lock, flags);
	if (!host) return -ENODEV;
        PDEBUG("ioctl-ret: host %d.%d.%d.%d (%08lx), count %i, to %i\n",
               NIPQUAD(info.ip), (long)info.mask,
	       info.bytes_per_second, info.timeout);
        copy_to_user(uptr, &info, sizeof(info));
        return 0;

      default:
        return -ENOTTY;
    }
    return 0;
}

/* ------------------------------------------------ The Network Device */

int rshaper_init_dev(struct net_device *dev)
{
    /* nothing more than ioctl */
    dev->do_ioctl = rshaper_ioctl;
    return 0;
}


struct net_device rshaper_dev = {
    name: "rshaper",
    init: rshaper_init_dev,
}; 


/* ------------------------------------------------ Module Stuff */

/* Register the initialization status */
static int init_status_flag;
   #define RSHAPER_RX_HOOK_REGISTERED		0x02
   #define RSHAPER_TX_HOOK_REGISTERED		0x04
   #define RSHAPER_NETDEVICE_REGISTERED		0x20

void cleanup_module(void)
{
    struct rshaper_host *host, *next;
    unsigned long flags;

/* de-hook */
#ifdef LINUX_24
    if (init_status_flag&RSHAPER_RX_HOOK_REGISTERED)   /* de-hook rx */
       nf_unregister_hook(&rshaper_rx_hook_ops);

    if (init_status_flag&RSHAPER_TX_HOOK_REGISTERED)   /* de-hook tx */
       nf_unregister_hook(&rshaper_tx_hook_ops);
#else
    net_shaper_rx_hook = NULL; /* de-hook */
#endif

    spin_lock_irqsave(&rshaper_host_lock, flags);
    while ( (host = first_host) ) { /* free it all */
        first_host = host->next;
        rshaper_kfree_host(host);
        host = next;
    }
    spin_unlock_irqrestore(&rshaper_host_lock, flags);

    if (init_status_flag&RSHAPER_NETDEVICE_REGISTERED)
        unregister_netdev(&rshaper_dev);
}

int init_module(void)
{
   int result;

    spin_lock_init(&rshaper_host_lock);
#ifdef LINUX_24
    PDEBUG("mode is %i\n", mode);

    /* Register the hook */
    if (mode==1 || mode==2) {
       result = nf_register_hook(&rshaper_rx_hook_ops);
       if (result < 0) {
           printk(KERN_ERR "rshaper: can't register netfilter hook");
           cleanup_module();
           return result;
       }
       init_status_flag |= RSHAPER_RX_HOOK_REGISTERED;
    }

    if (mode==0 || mode==2) {
       result = nf_register_hook(&rshaper_tx_hook_ops);
       if (result < 0) {
           printk(KERN_ERR "rshaper: can't register netfilter hook");
           cleanup_module();
           return result;
       }
       init_status_flag |= RSHAPER_TX_HOOK_REGISTERED;
    }
#else /* 2.2, 2.0 */
    net_shaper_rx_hook = rshaper_external_hook;
#endif

    result = register_netdev(&rshaper_dev);
    if (result) {
        printk(KERN_ERR "rshaper: can't register netdevice");
        cleanup_module();
        return result;
    }
    init_status_flag |= RSHAPER_NETDEVICE_REGISTERED;

    return 0; 
}

#ifndef LINUX_24 /* a one-argument hook calls the real hook */
int rshaper_external_hook(struct sk_buff *skb)
{
   return rshaper_rx_hook(0,		/* no hooknum */
                          &skb,         
                          NULL,         /* no in device */
                          NULL,         /* no out device */
                          (int (*)(struct sk_buff *)) netif_rx);
}


#endif

